@@ -22,7 +22,7 @@ import umap.distances as dist
 
 import umap.sparse as sparse
 
-from umap.utils import tau_rand_int, deheap_sort
+from umap.utils import tau_rand_int, deheap_sort, submatrix
 from umap.rp_tree import rptree_leaf_array, make_forest
 from umap.nndescent import (
     make_nn_descent,
@@ -367,27 +367,30 @@ def fuzzy_simplicial_set(
         returns a float can be provided. For performance purposes it is
         required that this be a numba jit'd function. Valid string metrics
         include:
-            * euclidean
-            * manhattan
-            * chebyshev
-            * minkowski
-            * canberra
+            * euclidean (or l2)
+            * manhattan (or l1)
+            * cityblock
             * braycurtis
-            * mahalanobis
-            * wminkowski
-            * seuclidean
-            * cosine
+            * canberra
+            * chebyshev
             * correlation
-            * haversine
+            * cosine
+            * dice
             * hamming
             * jaccard
-            * dice
-            * russelrao
             * kulsinski
+            * mahalanobis
+            * matching
+            * minkowski
             * rogerstanimoto
+            * russellrao
+            * seuclidean
             * sokalmichener
             * sokalsneath
+            * sqeuclidean
             * yule
+            * wminkowski
+
         Metrics that take arguments (such as minkowski, mahalanobis etc.)
         can have arguments passed via the metric_kwds dictionary. At this
         time care must be taken and dictionary elements must be ordered
@@ -501,8 +504,7 @@ def fast_intersection(rows, cols, values, target, unknown_dist=1.0, far_dist=5.0
 
     Returns
     -------
-    simplicial_set: sparse matrix
-        The resulting intersected fuzzy simplicial set.
+    None
     """
     for nz in range(rows.shape[0]):
         i = rows[nz]
@@ -677,7 +679,7 @@ def rdist(x, y):
     return result
 
 
-@numba.njit(fastmath=True)
+@numba.njit(fastmath=True, parallel=True)
 def optimize_layout(
     head_embedding,
     tail_embedding,
@@ -773,8 +775,11 @@ def optimize_layout(
 
                 dist_squared = rdist(current, other)
 
-                grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
-                grad_coeff /= a * pow(dist_squared, b) + 1.0
+                if dist_squared > 0.0:
+                    grad_coeff = -2.0 * a * b * pow(dist_squared, b - 1.0)
+                    grad_coeff /= a * pow(dist_squared, b) + 1.0
+                else:
+                    grad_coeff = 0.0
 
                 for d in range(dim):
                     grad_d = clip(grad_coeff * (current[d] - other[d]))
@@ -796,16 +801,19 @@ def optimize_layout(
 
                     dist_squared = rdist(current, other)
 
-                    grad_coeff = 2.0 * gamma * b
-                    grad_coeff /= (0.001 + dist_squared) * (
-                        a * pow(dist_squared, b) + 1
-                    )
-
-                    if not np.isfinite(grad_coeff):
-                        grad_coeff = 4.0
+                    if dist_squared > 0.0:
+                        grad_coeff = 2.0 * gamma * b
+                        grad_coeff /= (0.001 + dist_squared) * (
+                            a * pow(dist_squared, b) + 1
+                        )
+                    else:
+                        grad_coeff = 0.0
 
                     for d in range(dim):
-                        grad_d = clip(grad_coeff * (current[d] - other[d]))
+                        if grad_coeff > 0.0:
+                            grad_d = clip(grad_coeff * (current[d] - other[d]))
+                        else:
+                            grad_d = 4.0
                         current[d] += grad_d * alpha
 
                 epoch_of_next_negative_sample[i] += (
@@ -1206,7 +1214,7 @@ class UMAP(BaseEstimator):
         n_epochs=None,
         learning_rate=1.0,
         init="spectral",
-            min_dist=0.1,
+        min_dist=0.1,
         spread=1.0,
         set_op_mix_ratio=1.0,
         local_connectivity=1.0,
@@ -1228,18 +1236,11 @@ class UMAP(BaseEstimator):
 
         self.n_neighbors = n_neighbors
         self.metric = metric
-        if metric_kwds is not None:
-            self._metric_kwds = metric_kwds
-        else:
-            self._metric_kwds = {}
+        self.metric_kwds = metric_kwds
         self.n_epochs = n_epochs
-        if isinstance(init, np.ndarray):
-            self.init = check_array(init, dtype=np.float32, accept_sparse=False)
-        else:
-            self.init = init
+        self.init = init
         self.n_components = n_components
         self.repulsion_strength = repulsion_strength
-        self.initial_alpha = learning_rate
         self.learning_rate = learning_rate
 
         self.spread = spread
@@ -1252,26 +1253,14 @@ class UMAP(BaseEstimator):
         self.transform_queue_size = transform_queue_size
         self.target_n_neighbors = target_n_neighbors
         self.target_metric = target_metric
-        if target_metric_kwds is not None:
-            self._target_metric_kwds = target_metric_kwds
-        else:
-            self._target_metric_kwds = {}
+        self.target_metric_kwds = target_metric_kwds
         self.target_weight = target_weight
         self.transform_seed = transform_seed
         self.verbose = verbose
 
-        self._transform_available = False
+        self.a = a
+        self.b = b
 
-        self._validate_parameters()
-
-        if a is None or b is None:
-            self._a, self._b = find_ab_params(self.spread, self.min_dist)
-        else:
-            self._a = a
-            self._b = b
-
-        if self.verbose:
-            print(str(self))
 
     def _validate_parameters(self):
         if self.set_op_mix_ratio < 0.0 or self.set_op_mix_ratio > 1.0:
@@ -1279,9 +1268,9 @@ class UMAP(BaseEstimator):
         if self.repulsion_strength < 0.0:
             raise ValueError("repulsion_strength cannot be negative")
         if self.min_dist > self.spread:
-            raise ValueError("min_dist must be less than spread")
+            raise ValueError("min_dist must be less than or equal to spread")
         if self.min_dist < 0.0:
-            raise ValueError("min_dist must greater than 0.0")
+            raise ValueError("min_dist must be greater than 0.0")
         if not isinstance(self.init, str) and not isinstance(self.init, np.ndarray):
             raise ValueError("init must be a string or ndarray")
         if isinstance(self.init, str) and self.init not in ("spectral", "random"):
@@ -1295,7 +1284,7 @@ class UMAP(BaseEstimator):
             raise ValueError("metric must be string or callable")
         if self.negative_sample_rate < 0:
             raise ValueError("negative sample rate must be positive")
-        if self.initial_alpha < 0.0:
+        if self._initial_alpha < 0.0:
             raise ValueError("learning_rate must be positive")
         if self.n_neighbors < 2:
             raise ValueError("n_neighbors must be greater than 2")
@@ -1306,9 +1295,10 @@ class UMAP(BaseEstimator):
         if self.n_components < 1:
             raise ValueError("n_components must be greater than 0")
         if self.n_epochs is not None and (
-            self.n_epochs < 0 or not isinstance(self.n_epochs, int)
+            self.n_epochs <= 10 or not isinstance(self.n_epochs, int)
         ):
-            raise ValueError("n_epochs must be a positive integer")
+            raise ValueError("n_epochs must be a positive integer "
+                             "larger than 10")
 
     def fit(self, X, y=None):
         """Fit X into an embedded space.
@@ -1333,20 +1323,52 @@ class UMAP(BaseEstimator):
         X = check_array(X, dtype=np.float32, accept_sparse="csr")
         self._raw_data = X
 
+        # Handle all the optional arguments, setting default
+        if self.a is None or self.b is None:
+            self._a, self._b = find_ab_params(self.spread, self.min_dist)
+        else:
+            self._a = self.a
+            self._b = self.b
+
+        if self.metric_kwds is not None:
+            self._metric_kwds = self.metric_kwds
+        else:
+            self._metric_kwds = {}
+
+        if self.target_metric_kwds is not None:
+            self._target_metric_kwds = self.target_metric_kwds
+        else:
+            self._target_metric_kwds = {}
+
+        if isinstance(self.init, np.ndarray):
+            init = check_array(self.init, dtype=np.float32, accept_sparse=False)
+        else:
+            init = self.init
+
+        self._initial_alpha = self.learning_rate
+
+        self._validate_parameters()
+
+        if self.verbose:
+            print(str(self))
+
+        # Error check n_neighbors based on data size
         if X.shape[0] <= self.n_neighbors:
             if X.shape[0] == 1:
-                return np.array([0.0, 0.0])  # needed to sklearn comparability
+                self.embedding_ = np.zeros((1, self.n_components))  # needed to sklearn comparability
+                return self
 
             warn(
                 "n_neighbors is larger than the dataset size; truncating to "
                 "X.shape[0] - 1"
             )
-            n_neighbors = X.shape[0] - 1
+            self._n_neighbors = X.shape[0] - 1
         else:
-            n_neighbors = self.n_neighbors
+            self._n_neighbors = self.n_neighbors
 
-        if scipy.sparse.isspmatrix_csr(X) and not X.has_sorted_indices:
-            X.sort_indices()
+        if scipy.sparse.isspmatrix_csr(X):
+            if not X.has_sorted_indices:
+                X.sort_indices()
             self._sparse_data = True
         else:
             self._sparse_data = False
@@ -1362,7 +1384,7 @@ class UMAP(BaseEstimator):
             dmat = pairwise_distances(X, metric=self.metric, **self._metric_kwds)
             self.graph_ = fuzzy_simplicial_set(
                 dmat,
-                n_neighbors,
+                self._n_neighbors,
                 random_state,
                 "precomputed",
                 self._metric_kwds,
@@ -1378,7 +1400,7 @@ class UMAP(BaseEstimator):
             # Standard case
             (self._knn_indices, self._knn_dists, self._rp_forest) = nearest_neighbors(
                 X,
-                n_neighbors,
+                self._n_neighbors,
                 self.metric,
                 self._metric_kwds,
                 self.angular_rp_forest,
@@ -1400,8 +1422,6 @@ class UMAP(BaseEstimator):
                 self.verbose,
             )
 
-            self._raw_data = X
-            self._transform_available = True
             self._search_graph = scipy.sparse.lil_matrix(
                 (X.shape[0], X.shape[0]), dtype=np.int8
             )
@@ -1433,33 +1453,54 @@ class UMAP(BaseEstimator):
                 )
 
         if y is not None:
+            y_ = check_array(y, ensure_2d=False)
             if self.target_metric == "categorical":
                 if self.target_weight < 1.0:
                     far_dist = 2.5 * (1.0 / (1.0 - self.target_weight))
                 else:
                     far_dist = 1.0e12
                 self.graph_ = categorical_simplicial_set_intersection(
-                    self.graph_, y, far_dist=far_dist
+                    self.graph_, y_, far_dist=far_dist
                 )
             else:
                 if self.target_n_neighbors == -1:
-                    target_n_neighbors = self.n_neighbors
+                    target_n_neighbors = self._n_neighbors
                 else:
                     target_n_neighbors = self.target_n_neighbors
 
-                target_graph = fuzzy_simplicial_set(
-                    y[np.newaxis, :].T,
-                    target_n_neighbors,
-                    random_state,
-                    self.target_metric,
-                    self._target_metric_kwds,
-                    None,
-                    None,
-                    False,
-                    1.0,
-                    1.0,
-                    False,
-                )
+                # Handle the small case as precomputed as before
+                if y.shape[0] < 4096:
+                    ydmat = pairwise_distances(y_[np.newaxis, :].T,
+                                               metric=self.target_metric,
+                                               **self._target_metric_kwds)
+                    target_graph = fuzzy_simplicial_set(
+                        ydmat,
+                        target_n_neighbors,
+                        random_state,
+                        "precomputed",
+                        self._target_metric_kwds,
+                        None,
+                        None,
+                        False,
+                        1.0,
+                        1.0,
+                        False
+                    )
+                else:
+                    # Standard case
+                    target_graph = fuzzy_simplicial_set(
+                        y_[np.newaxis, :].T,
+                        target_n_neighbors,
+                        random_state,
+                        self.target_metric,
+                        self._target_metric_kwds,
+                        None,
+                        None,
+                        False,
+                        1.0,
+                        1.0,
+                        False,
+                    )
                 # product = self.graph_.multiply(target_graph)
                 # # self.graph_ = 0.99 * product + 0.01 * (self.graph_ +
                 # #                                        target_graph -
@@ -1482,13 +1523,13 @@ class UMAP(BaseEstimator):
             self._raw_data,
             self.graph_,
             self.n_components,
-            self.initial_alpha,
+            self._initial_alpha,
             self._a,
             self._b,
             self.repulsion_strength,
             self.negative_sample_rate,
             n_epochs,
-            self.init,
+            init,
             random_state,
             self.metric,
             self._metric_kwds,
@@ -1537,6 +1578,10 @@ class UMAP(BaseEstimator):
         X_new : array, shape (n_samples, n_components)
             Embedding of the new data in low-dimensional space.
         """
+        # If we fit just a single instance then error
+        if self.embedding_.shape[0] == 1:
+            raise ValueError('Transform unavailable when model was fit with'
+                             'only a single data sample.')
         # If we just have the original input then short circuit things
         X = check_array(X, dtype=np.float32, accept_sparse="csr")
         x_hash = joblib.hash(X)
@@ -1557,16 +1602,19 @@ class UMAP(BaseEstimator):
             dmat = pairwise_distances(
                 X, self._raw_data, metric=self.metric, **self._metric_kwds
             )
-            indices = np.argsort(dmat)
-            dists = np.sort(dmat)  # TODO: more efficient approach
-            indices = indices[:, : self.n_neighbors]
-            dists = dists[:, : self.n_neighbors]
+            indices = np.argpartition(dmat,
+                                      self._n_neighbors)[:, :self._n_neighbors]
+            dmat_shortened = submatrix(dmat, indices, self._n_neighbors)
+            indices_sorted = np.argsort(dmat_shortened)
+            indices = submatrix(indices, indices_sorted, self._n_neighbors)
+            dists = submatrix(dmat_shortened, indices_sorted,
+                              self._n_neighbors)
         else:
             init = initialise_search(
                 self._rp_forest,
                 self._raw_data,
                 X,
-                int(self.n_neighbors * self.transform_queue_size),
+                int(self._n_neighbors * self.transform_queue_size),
                 self._random_init,
                 self._tree_init,
                 rng_state,
@@ -1580,11 +1628,12 @@ class UMAP(BaseEstimator):
             )
 
             indices, dists = deheap_sort(result)
-            indices = indices[:, : self.n_neighbors]
-            dists = dists[:, : self.n_neighbors]
+            indices = indices[:, : self._n_neighbors]
+            dists = dists[:, : self._n_neighbors]
 
+        adjusted_local_connectivity = max(0, self.local_connectivity - 1.0)
         sigmas, rhos = smooth_knn_dist(
-            indices, self.n_neighbors, local_connectivity=self.local_connectivity
+            dists, self._n_neighbors, local_connectivity=adjusted_local_connectivity
         )
 
         rows, cols, vals = compute_membership_strengths(indices, dists, sigmas, rhos)
@@ -1593,9 +1642,12 @@ class UMAP(BaseEstimator):
             (vals, (rows, cols)), shape=(X.shape[0], self._raw_data.shape[0])
         )
 
+        # This was a very specially constructed graph with constant degree.
+        # That lets us do fancy unpacking by reshaping the csr matrix indices
+        # and data. Doing so relies on the constant degree assumption!
         csr_graph = normalize(graph.tocsr(), norm="l1")
-        inds = csr_graph.indices.reshape(X.shape[0], self.n_neighbors)
-        weights = csr_graph.data.reshape(X.shape[0], self.n_neighbors)
+        inds = csr_graph.indices.reshape(X.shape[0], self._n_neighbors)
+        weights = csr_graph.data.reshape(X.shape[0], self._n_neighbors)
         embedding = init_transform(inds, weights, self.embedding_)
 
         if self.n_epochs is None:
@@ -1627,7 +1679,7 @@ class UMAP(BaseEstimator):
             self._b,
             rng_state,
             self.repulsion_strength,
-            self.initial_alpha,
+            self._initial_alpha,
             self.negative_sample_rate,
             verbose=self.verbose,
         )
